/*
 * A user-configured filesystem.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>
#include <l4/util/util.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ext2fs/ext2fs.h>
#include <e2access/access.h>
#include <e2access/image.h>
#include <e2access/path.h>

#include <fsserver/file_resource.h>
#include <systypes/fcntl.h>
#include "fs_accessor.h"
#include "fs_directory.h"
#include "fs_directory_listing_accessor.h"
#include "fs_user_filesystem.h"



/* Check access to a file. */

long Fs_user_filesystem::_check_access(int flags, ext2_ino_t ino_file)
{
  struct ext2_inode inode_file;

  /* Check the file metadata for the user. */

  if (ext2fs_read_inode(_fs, ino_file, &inode_file))
    return -L4_EIO;

  if ((flags & O_RDWR) || (flags & O_WRONLY))
  {
    if (!access_can_write(_user, &inode_file))
      return -L4_EPERM;
  }
  else if (!access_can_read(_user, &inode_file))
    return -L4_EPERM;

  return L4_EOK;
}

/* Create a file. */

long Fs_user_filesystem::_create(const char *path, ext2_ino_t ino_parent,
                                 ext2_ino_t *ino_file)
{
  struct ext2_inode inode_parent;

  /* Check the directory metadata for the user. */

  if (ext2fs_read_inode(_fs, ino_parent, &inode_parent))
    return -L4_EIO;

  if (!access_can_write(_user, &inode_parent))
    return -L4_EPERM;

  /* Obtain a directory accessor. */

  Fs_directory_listing_accessor *accessor = dynamic_cast<Fs_directory_listing_accessor *>(_get_accessor(ino_parent));

  if (!accessor)
    return -L4_EIO;

  return accessor->create_file(path, _user.uid, _user.gid, 0666 & ~_user.umask, ino_file);
}

/* Find the given path in the filesystem. */

long Fs_user_filesystem::_find(const char *path, int flags, int *ext2flags,
                               ext2_ino_t *ino_file)
{
  ext2_ino_t ino_parent;
  long err;

  /* If the path does not yield a filesystem object... */

  if (image_find_path(_fs, &path, &ino_parent))
  {
    /* Test if a file can be created. */

    if ((flags & O_CREAT) && (flags & ~O_DIRECTORY))
    {
      *ext2flags |= EXT2_FILE_CREATE;

      err = _create(path, ino_parent, ino_file);
      if (err)
        return err;
    }
    else
      return -L4_EIO;
  }

  /* If the filesystem object exists... */

  else
  {
    *ino_file = ino_parent;

    err = _check_access(flags, *ino_file);
    if (err)
      return err;
  }

  /* Set the appropriate ext2fs flags. */

  if ((flags & O_RDWR) || (flags & O_WRONLY))
    *ext2flags |= EXT2_FILE_WRITE;

  return L4_EOK;
}

/* Open an accessor for a file or directory. */

Accessor *Fs_user_filesystem::_get_accessor(ext2_ino_t ino_file, int flags)
{
  Accessor *accessor, *found_accessor;

  if (_image_isfile(_fs, ino_file))
    accessor = new Fs_accessor(_fs, ino_file, flags, _paging);
  else if (_image_isdir(_fs, ino_file))
    accessor = new Fs_directory_listing_accessor(_fs, ino_file, flags, _paging);
  else
    return 0;

  found_accessor = _paging->set_conditional(ino_file, accessor);

  if (found_accessor != accessor)
  {
    delete accessor;
    printf("Found accessor: %p\n", found_accessor);
  }
  else
  {
    found_accessor = accessor;
    printf("Made new accessor: %p\n", accessor);
  }

  return found_accessor;
}

/* Create a resource for a file or directory. */

Resource *Fs_user_filesystem::_get_resource(ext2_ino_t ino_file, fs_object_t *fsobj, Accessor *accessor)
{
  /* Populate the size member of the common object structure.
     Create a new resource initialised with an accessor. */

  if (_image_isfile(_fs, ino_file))
  {
    PagingAccessor *acc = dynamic_cast<PagingAccessor *>(accessor);
    fsobj->size = acc->get_size();
    return new MappedFileResource(acc, *fsobj);
  }
  else if (_image_isdir(_fs, ino_file))
  {
    PagingAccessor *acc = dynamic_cast<PagingAccessor *>(accessor);
    fsobj->size = acc->get_size();
    return new Fs_directory(acc, *fsobj, _fs, ino_file);
  }
  else
    return NULL;
}

/* Obtain a resource server for a filesystem object. */

long Fs_user_filesystem::open_object_resource(fs_object_t *fsobj, int flags,
                                              Resource **resource)
{
  int ext2flags = 0;
  ext2_ino_t ino_file;
  long err;

  /* Obtain the object's path. */

  const char *path = (const char *) fsdesc_get_name(&fsobj->buffer);

  if (path == NULL)
    return -L4_EINVAL;

  /* Find the object. If it cannot be located, return an error if it is not to
     be created. */

  err = _find(path, flags, &ext2flags, &ino_file);

  if (err)
    return err;

  /* Look up any existing objects for the file. */

  Accessor *accessor = _get_accessor(ino_file, ext2flags);

  /* Send an error for unsupported objects. */

  if (!accessor)
    return -L4_EIO;

  *resource = _get_resource(ino_file, fsobj, accessor);

  if (*resource == NULL)
    return -L4_EIO;

  return L4_EOK;
}
