/*
 * Access an ext2-based filesystem.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>

#include <stdio.h>
#include <stdlib.h>

#include <ext2fs/ext2fs.h>
#include <fsserver/resource_server.h>

#include "fs_filesystem.h"



/* Access to block servers for libext2fs. */ 

extern io_manager blockserver_io_manager;



int main(int argc, char *argv[])
{
  char *devname;
  errcode_t retval;
  int num_pages;

  if (argc < 3)
  {
    printf("Need a block device capability or a file accessible via the fs capability.\n"
           "Also need a number of pages to reserve for data.\n");
    return 1;
  }

  devname = argv[1];
  num_pages = atoi(argv[2]);

  /* Obtain access to a filesystem situated on a block device. */

  ext2_filsys fs = NULL;
  int ext2flags = EXT2_FLAG_RW; // | EXT2_FLAG_SOFTSUPP_FEATURES | EXT2_FLAG_64BITS;

  if ((retval = ext2fs_open(devname, ext2flags, 0, 0, blockserver_io_manager, &fs)))
  {
    printf("Could not obtain filesystem: %ld.\n", retval);
    return 1;
  }

  if ((retval = ext2fs_read_bitmaps(fs)))
  {
    printf("Could not obtain filesystem: %ld.\n", retval);
    return 1;
  }

  /* Allocate pages for filesystem data. */

  Paging *paging = new Paging(new Pages(L4_PAGESIZE * num_pages));

  /* Initialise and register a new server object. */

  Fs_filesystem filesystem(paging, fs, devname);
  ResourceServer server(&filesystem);

  if (server.bind("export"))
  {
    printf("Could not bind thread.\n");
    return 1;
  }

  /* Enter the IPC server loop. */

  printf("Starting e2fsserver for %s...\n", devname);
  server.start();
  return 0;
}
